// PState.m
//
//----------------------------------------------------------------
// OOF Parser Package
// Copyright 1995 by Paul Burchard
// Distributed according to the GNU Library General Public License
//----------------------------------------------------------------

#import <string.h>
#import <objects/Stack.h>
#import <objects/Dictionary.h>
#import "BootstrapParsers.h"
#import "BuiltinParsers.h"
#import "ParserParsers.h"
#import "ParserAutoreleasePool.h"
#import "PState.h"
#import "alloc.h"
#import "elt_p.h"

#define RETURN_BY_CALLING_EXCEPTION_FUNCTION(f) return (*(f))(__builtin_apply_args())

typedef struct parser_state_state
{
	void *stack_state;
	id scope;
	void *scope_state;
}
parser_state_state;

@implementation PState
- appendBuiltins
{
	[self putObject:[[IgnorableParser new] autorelease] atName:"ignorable"];
	[self putObject:[[LiteralParser new] autorelease] atName:"literal"];
	[self putObject:[[IntegerParser new] autorelease] atName:"integer"];
	[self putObject:[[SymbolParser new] autorelease] atName:"symbol"];
	[self putObject:[[EndParser new] autorelease] atName:"end"];

	[self putObject:(id)[ArgList class] atName:"Default"];

	[self putObject:[[BNFListParser new] autorelease] atName:"list"];
	[self putObject:[[BNFParser new] autorelease] atName:"bnf"];
	[self putObject:[[BNFStatementParser new] autorelease] atName:"bnf_statement"];
	[self putObject:[[BNFParserParser new] autorelease] atName:"bnf_parser"];
	[self putObject:[[BNFTermParser new] autorelease] atName:"bnf_term"];
	[self putObject:[[BNFFactorParser new] autorelease] atName:"bnf_factor"];
	[self putObject:[[BNFAtomParser new] autorelease] atName:"bnf_atom"];
	[self putObject:[[BNFSuffixParser new] autorelease] atName:"bnf_suffix"];
	return self;
}
- init
{
	return [self initWithType:@encode(id) keyType:ENCODE_MUTABLE_STRING];
}
- initWithType:(const char *)contentEncoding keyType:(const char *)keyEncoding
{
	id symbols = nil;
	
	if((contentEncoding && !ENCODING_IS_OBJECT(contentEncoding))
		|| (keyEncoding && !ENCODING_IS_STRING(keyEncoding)))
			{ [self release]; return nil; }
	if(![super initWithType:@encode(id) keyType:ENCODE_MUTABLE_STRING]) return nil;
	if(!(symbols = [[Dictionary alloc] initWithType:@encode(id) keyType:ENCODE_MUTABLE_STRING])
		|| !(stack = [[Stack alloc] initWithCapacity:10]))
			{ [symbols release]; [self release]; return nil; }
	[stack pushObject:[symbols autorelease]];
	[self appendBuiltins];
	return self;
}
- (void)dealloc
{
	[stack release];
	[super dealloc];
}
- objectAtName:(const char *)aName;
{
	// Just return nil if absent.
	return [self elementAtKey:ELT_STR(aName) ifAbsentCall:NILXF].id_u;
}
- putObject:anObj atName:(const char *)aName
{
	return [self putElement:ELT(id, anObj) atKey:ELT_STR(aName)];
}
- (const char *)contentType
{
	return @encode(id);
}
- (const char *)keyType
{
	return ENCODE_MUTABLE_STRING;
}
- (BOOL)includesKey:(elt)aKey
{
	id cls;
	BOOL notFound = YES;
	void checkKey(id scope)
	{
		if([scope includesKey:aKey]) notFound = NO;
	}
	[stack withObjectsInReverseCall:checkKey whileTrue:&notFound];
	if(notFound && aKey.char_ptr_u
		&& (cls = (id)objc_lookup_class(aKey.char_ptr_u))
		&& [cls conformsToProtocol:@protocol(Parsing)])
			notFound = NO;
	return !notFound;	
}
- (elt)elementAtKey:(elt)aKey ifAbsentCall:(elt(*)(arglist_t))excFunc
{
	elt value;
	BOOL notFound = YES;
	void checkObj(id scope)
	{
		if( (value = [scope elementAtKey:aKey ifAbsentCall:NILXF]).id_u )
			notFound = NO;
	}
	value = ELT(id, nil);
	[stack withObjectsInReverseCall:checkObj whileTrue:&notFound];
	if(notFound && aKey.char_ptr_u
		&& (value.id_u = (id)objc_lookup_class(aKey.char_ptr_u))
		&& [value.id_u conformsToProtocol:@protocol(Parsing)])
			notFound = NO;
	if(notFound) RETURN_BY_CALLING_EXCEPTION_FUNCTION(excFunc);
	return value;	
}

- putElement:(elt)newContentElement atKey:(elt)aKey
{
	// Binding is placed in innermost (current) scope.
	if(![[stack topObject] putElement:newContentElement atKey:aKey])
		return nil;

	// Cache scoped input.
	if(aKey.char_ptr_u && 0==strcmp(aKey.char_ptr_u, "<"))
		input = newContentElement.id_u;
	return self;
}
- (elt)removeElementAtKey:(elt)aKey ifAbsentCall:(elt(*)(arglist_t))excFunc
{
	// Can only remove from current scope.
	elt value = [[stack topObject] removeElementAtKey:aKey ifAbsentCall:excFunc];

	// Cache scoped input.
	input = [self elementAtKey:ELT_STR("<") ifAbsentCall:NILXF].id_u;

	return value;
}
- (BOOL)getNextKey:(elt*)aKeyPtr content:(elt*)anElementPtr withEnumState:(void**)enumState
{
	// Iterate through each scope in turn, moving outwards (=downwards).
	parser_state_state **state = (parser_state_state **)enumState;
	if(!state) return NO;
	if(!*state)
	{
		if(!ALLOC(*state, struct parser_state_state, 1)) return NO;
		(*state)->stack_state = (*state)->scope_state = 0;
		(*state)->scope = nil;
	}
	if(!(*state)->scope || ![(*state)->scope getNextKey:aKeyPtr
		content:anElementPtr withEnumState:&(*state)->scope_state])
	{
		if(![stack getPrevObject:&(*state)->scope
			withEnumState:&(*state)->stack_state])
				{ FREE(*state); return NO; }
	}
	return YES;
}
- (unsigned)count
{
	unsigned n = 0;
	BOOL always = YES;
	void accumCount(id scope)
	{
		n += [scope count];
	}
	[stack withObjectsInReverseCall:accumCount whileTrue:&always];
	return n;	
}
- empty
{
	// Can only empty current scope.
	id ok = [[stack topObject] empty];

	// Cache scoped input.
	input = [self elementAtKey:ELT_STR("<") ifAbsentCall:NILXF].id_u;

	return ok;
}
- input
{
	return input;
}
- push
{
	// Push new empty scope onto stack.
	// If this is bottom of stack, add built-in symbols.
	id symbols;
	if(!(symbols = [[Dictionary alloc] initWithType:@encode(id) keyType:ENCODE_MUTABLE_STRING])) return nil; //!!!
	if([stack count] == 0) [self appendBuiltins];
	[stack pushObject:[symbols autorelease]];

	// Create new autorelease pool for this scope.
	[[ParserAutoreleasePool alloc] init];
	return self;
}
- pop
{
	// Pop off top scope (autoreleased).
	id ok = self;
	if([stack count] <= 1) return nil;
	if(![stack popObject]) ok = nil;
	
	// Execute pending autoreleases for scope just closed.
	[[ParserAutoreleasePool currentPool] release];

	// Cache new value of scoped input.
	input = [self elementAtKey:ELT_STR("<") ifAbsentCall:NILXF].id_u;
	return ok;
}
@end



