// ParserParsers.m
//
//----------------------------------------------------------------
// OOF Parser Package
// Copyright 1995 by Paul Burchard
// Distributed according to the GNU Library General Public License
//----------------------------------------------------------------

#import <stdio.h>
#import <string.h>
#import <ctype.h>
#import <objc/objc-api.h>
#import <objects/Stream.h>
#import <objects/StdioStream.h>
#import <objects/String.h>
#import "ParserStateCollecting.h"
#import "ParserAutoreleasePool.h"
#import "ParserParsers.h"
#import "PValue.h"
#import "alloc.h"
#import "elt_p.h"


@implementation ArgList
- (BOOL)isFinalized
{
	int i, n;
	n = [self count];
	for(i=0; i<n; i++) if(![[self objectAtIndex:i] isFinalized]) return NO;
	return YES;
}
- finalizeWith:state
{
	// Finalize each of the args in turn.
	int i = 0;
	if([self isFinalized]) return self;
	while(i < [self count])
	{
		id arg = [[self removeObjectAtIndex:i] finalizeWith:state];
		if(arg) [self insertObject:arg atIndex:i++];
	}
	return self;
}
- (const char *)evalType { return @encode(id); }
- (const char *)evalElementType
{
	if([self count] > 0) return [[self objectAtIndex:0] evalType];
	else return @encode(id);
}
- (elt)evalWith:state
{
	// Return list of values of args (C type must be uniform).
	// Subclasses normally override this completely.
	// Return value is new object (or old one w/ retain count increased).
	int i, n = [self count];
	const char *valEltType = [self evalElementType];
	id value = [[ArgList alloc] initWithType:valEltType capacity:0];
	if(!value) return ELT(id, nil);
	for(i=0; i<n; i++)
		[value appendElement:[[self objectAtIndex:i] evalWith:state]];
	return ELT(id, [value autorelease]);
}
- init
{
	return [self initWithType:@encode(id) capacity:0];
}
//!!! Should also override +setAutoreleaseClass: and
//!!! +autoreleaseClass in newer versions of libobjects.
- autorelease
{
	[ParserAutoreleasePool addObject:self];
	return self;
}
- copy
{
	// No [super deepenCopy] in Collection hierarchy.
	return [[super copy] deepenCopy];
}
- deepenCopy
{
	// No [super deepenCopy] in Collection hierarchy.
	return self;
}
- defaultConstructorWith:state
{
	// Default constructor must be capable of being used as
	// a generic list (i.e., unfinalized).
	id ctor = [state objectAtName:"Default"];
	if(!ctor) ctor = (id)[ArgList class];
	return ctor;
}
- removeNonParsing
{
	int i;
	id obj;
	
	// Get rid of elements that are not <Parsing> or return non-objects.
	for(i=(int)[self count]-1; i>=0; i--)
	{
		obj = [self objectAtIndex:i];
		if(![obj conformsToProtocol:@protocol(Parsing)] || !ENCODING_IS_OBJECT([obj evalType]))
			[self removeObjectAtIndex:i];
	}
	return self;
}
- traceObject:anObj with:state
{
	//!!! Note Stream classes in libobjects may change.
	id output;
	if(anObj && (output = [state objectAtName:"2>"])
		&& [output respondsToSelector:@selector(writeFormat:)])
	{
		int i, c;
		const char *s;
		id input;
		elt mark;

		if([anObj respondsToSelector:@selector(printLispTo:)])
		{
			[output writeFormat:"\nVALUE: "];
			[anObj printLispTo:output];
		}
		else if(getStringValue(anObj, &s))
			[output writeFormat:"\nVALUE: {%s \"%s\"}", [anObj name], s];
		else if(getIntValue(anObj, &i))
			[output writeFormat:"\nVALUE: {%s %d}", [anObj name], i];
		
		if([self respondsToSelector:@selector(printLispTo:)])
		{
			[output writeFormat:"\nFUNCTION: "];
			[self printLispTo:output];
		}

		input = [state input];
		mark = [input nextKey];
		[output writeFormat:"\nINPUT REMAINING: \""];
		for(i=0, c=[input elementAtNextKey].int_u; i<20 && c!=EOF; i++, c=[input elementAtNextKey].int_u)
			[output writeFormat:"%c", c];
		if(c != EOF) [output writeFormat:"\"..."];
		else [output writeFormat:"\""];
		[input setNextKey:mark];

		[output writeFormat:"\n"];
		if([output respondsToSelector:@selector(flushStream)])
			[output flushStream];
	}
	return anObj;
}
- printLispTo:stream
{
	//!!! Note: -superClass should change to -superclass in GNU.
	//!!! Note: -class method is screwed up because of Class/Class* conflict.
	//!!!       The (Object *) casts should be removed when libobjects is fixed.
	//!!! Note: Stream classes are in flux in libobjects
	//!!! Note: -name is not in OpenStep.
	int i, k;
	id cls;
	const char *s;
	if(![stream respondsToSelector:@selector(writeFormat:)]) return nil;
	[stream writeFormat:"({"];
	if((id)[(Object *)self class] == (id)[ArgList class])
		[stream writeFormat:"%s", [(id)[(Object *)self class] name]];
	else for(cls=(id)[(Object *)self class];
		cls!=(id)[cls superClass] && cls!=(id)[ArgList class];
		cls=(id)[cls superClass])
			[stream writeFormat:"%s%s",
				(cls==(id)[(Object *)self class]?"":":"),
				[cls name]];
	if([self respondsToSelector:@selector(printIvarsTo:)])
		[self printIvarsTo:stream];
	[stream writeFormat:"}"];
	if([self respondsToSelector:@selector(count)]) for(i=0; i<[self count]; i++)
	{
		[stream writeFormat:" "];
		if([[self objectAtIndex:i] respondsToSelector:@selector(printLispTo:)])
			[[self objectAtIndex:i] printLispTo:stream];
		else if(getStringValue([self objectAtIndex:i], &s))
			[stream writeFormat:"{%s \"%s\"}", [[self objectAtIndex:i] name], s];
		else if(getIntValue([self objectAtIndex:i], &k))
			[stream writeFormat:"{%s %d}", [[self objectAtIndex:i] name], k];
		else
			[stream writeFormat:"{%s}", [[self objectAtIndex:i] name]];
	}
	[stream writeFormat:")"];
	return self;
}
- printIvarsTo:stream
{
	// Subclasses fill this in.
	return self;
}
@end

@implementation ParserSequence
- (BOOL)isFinalized
{
	if(!isFinalSeq) return NO;
	return [super isFinalized];
}
- (void)setFinalized:(BOOL)flag
{
	isFinalSeq = flag;
}
- finalizeWith:state
{
	// Clear out non-functions from arg list.
	// As a special case, if this leaves exactly one function (and we
	// use default constructor) replace ourselves with this function.
	id replacement;
	if([self isFinalized]) return self;
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	if(!isFinalSeq)
	{
		[self retain];
		[self removeNonParsing];
		[self autorelease];
		isFinalSeq = YES;
	}
	if([self count]==1 && (![self cStringValue] || 0==strcmp([self cStringValue], "Default")))
		return [self removeObjectAtIndex:0];
	else return self;
}
- (elt)evalWith:state
{
	// Build a function object of the constructor class,
	// with the sequence of parser results as syntax arguments.
	// If we are an immediate constructor, finalize after construction
	// and clear input buffer.  Note the constructor name is scoped,
	// and resolved at run time.
	elt mark;
	id arg, input, value, ctor;
	int i, n = [self count];
	const char *ctorName = [self cStringValue];
	if(!ctorName || !(ctor = [state objectAtName:ctorName]))
		ctor = [self defaultConstructorWith:state];
	if(!(value = [ctor new])) return ELT(id, nil);
	input = [state input];
	mark = [input nextKey];
	for(i=0; i<n; i++)
	{
		if( (arg = [[self objectAtIndex:i] evalWith:state].id_u) )
			[value appendObject:arg];
		else
		{
			[value release];
			[input setNextKey:mark];
			return ELT(id, nil);
		}
	}
	if(isImmediate)
	{
		value = [[value finalizeWith:state] retain];
		[input removeBeforeKey:[input nextKey]];
	}
	return ELT(id, [[self traceObject:value with:state] autorelease]);
}
- (const char *)cStringValue { return 0; }
- (BOOL)isImmediate { return isImmediate; }
- (void)setImmediate:(BOOL)flag { isImmediate = flag; }
@end

@implementation ParserDefault
- (BOOL)isFinalized
{
	if(!isFinalMerge) return NO;
	return [super isFinalized];
}
- (void)setFinalized:(BOOL)flag
{
	isFinalMerge = flag;
	[super setFinalized:flag];
}
- finalizeWith:state
{
	// Sole (or first) syntax arg should be generic list which bundles
	// the real Sequence syntax args; replace our args with this list.
	// (Do not finalize merged args; leave that up to superclass.)
	// As a special case, if this leaves exactly one function (and we
	// use default constructor) replace ourselves with this function.
	id args = nil;
	if([self isFinalized]) return self;
	[self retain];
	if([self count] > 0) args = [self removeObjectAtIndex:0];
	[self empty];
	if(args) [self appendContentsOf:args];
	[self autorelease];
	isFinalMerge = YES;
	return [super finalizeWith:state];
}
@end

@implementation ParserConstructor
- (BOOL)isFinalized
{
	if(!constructorName) return NO;
	return [super isFinalized];
}
- finalizeWith:state
{
	// First syntax arg is constructor name; second syntax arg is a
	// literal indicating whether the construction should occur
	// immediately upon recognition by the parser.  The remaining
	// function syntax args are treated as for ParserDefault.
	const char *newName = 0;
	if([self isFinalized]) return self;
	if(!constructorName && getStringValue([[self removeObjectAtIndex:0]
		finalizeWith:state], &newName) && newName)
	{
		[self setCStringValue:newName];
		newName = 0;
		if(getStringValue([[self removeObjectAtIndex:0]
			finalizeWith:state], &newName) && newName)
				if(newName[0] == '!') isImmediate = YES;
	}
	return [super finalizeWith:state];
}
- initWithCString:(const char *)ctorName
{
	if(![self initWithType:@encode(id) capacity:2]) return nil;
	[self setCStringValue:ctorName];
	return self;
}
- (void)dealloc { FREE(constructorName); [super dealloc]; }
- deepenCopy
{
	if(![super deepenCopy]) return nil;
	if(constructorName) constructorName = strdup(constructorName);
	return self;
}
- (const char *)cStringValue { return constructorName; }
- (void)setCStringValue:(const char *)ctorName
{
	FREE(constructorName);
	if(ctorName) constructorName = strdup(ctorName);
}
- printIvarsTo:stream
{
	[super printIvarsTo:stream];
	if([self cStringValue]) [stream writeFormat:" <%s>%s", [self cStringValue], (isImmediate?"!":"")];
	else [stream writeFormat:" <Default>%s", (isImmediate?"!":"")];
	return self;
}
@end

@implementation ParserSuffix
- (BOOL)isFinalized
{
	// Exists only to append suffixes and die.
	return NO;
}
- finalizeWith:state
{
	// ParserSuffix is similar to ParserConstructor, but it
	// gets constructor info, and possibly some extra args,
	// from its final "suffix" syntax arg instead of from itself.
	// Also, as a result, the actual construction takes place at
	// finalization time instead of eval time.  (Immediacy is thus
	// controlled by immediacy of the construction of the ParserSuffix.)
	//
	// Do not finalize the suffix in advance; but do finalize
	// and return the entire constructed function.
	const char *ctorName = 0;
	id suffix, ctor, replacement;
	if([self count]<1
		|| !(suffix = [self removeLastObject])
		|| ![suffix conformsToProtocol:@protocol(Parsing)]
		|| ![suffix respondsToSelector:@selector(cStringValue)])
			return nil;
	if(!(ctorName = [suffix cStringValue]) || !(ctor = [state objectAtName:ctorName]))
		ctor = [self defaultConstructorWith:state];
	if(!(replacement = [ctor new])) return nil;
	[replacement appendContentsOf:self];
	[replacement appendContentsOf:suffix];
	return [[replacement autorelease] finalizeWith:state];
}
- (elt)evalWith:state
{
	// Exists only to append suffixes and die.
	return ELT(id, nil);
}
@end

@implementation SuffixParser
- (BOOL)isFinalized
{
	// Exists only to hold suffix for ParserSuffix and be consumed.
	return NO;
}
- finalizeWith:state
{
	// Arguments should be held unfinalized until inserted into
	// eventual constructor.
	return self;
}
- (elt)evalWith:state
{
	// Exists only to hold suffix for ParserSuffix and be consumed.
	return ELT(id, nil);
}
- (const char *)cStringValue
{
	// Subclass responsibility to override.
	return 0;
}
@end

@implementation SuffixComposition
- (const char *)cStringValue { return "ParserComposition"; }
@end

@implementation SuffixRepetition
- (const char *)cStringValue { return "ParserRepetition"; }
@end

@implementation SuffixSequence
- (const char *)cStringValue { return "ParserSequence"; }
@end

@implementation ParserRepetition
- finalizeWith:state
{
	int i, numValue = (-1);
	const char *ineqValue = 0;
	id replacement, bounds = nil, num = nil, ineq = nil;

	// Last syntax argument can be one of the following:
	//   1. A number.
	//   2. An optional number.
	//   3. An optional sequence of an optional inequality and a number.
	// There must be at least one additional syntax argument.
	if([self count]<2 || !(bounds = [[self removeLastObject] finalizeWith:state])) { num = ineq = nil; }
	else if(![bounds conformsToProtocol:@protocol(Parsing)]) { num = bounds; ineq = nil; }
	else if([bounds count]<1 || !(bounds = [bounds objectAtIndex:0])) { num = ineq = nil; }
	else if(![bounds conformsToProtocol:@protocol(Parsing)]) { num = bounds; ineq = nil; }
	else if([bounds count]<2 || ![[bounds objectAtIndex:0] conformsToProtocol:@protocol(Parsing)]) { num = ineq = nil; }
	else if([[bounds objectAtIndex:0] count] < 1) { num = [bounds lastObject]; ineq = nil; }
	else { num = [bounds lastObject]; ineq = [[bounds objectAtIndex:0] objectAtIndex:0]; }

	// Set bounds according to spec, if present.
	if(!num || !getIntValue(num, &numValue))
		/* leave defaults (unbounded repetition) */;
	else if(!ineq || !getStringValue(ineq, &ineqValue) || !ineqValue)
		[self setLowerBound:numValue upperBound:numValue];
	else if(0 == strcmp(ineqValue, ">"))
		[self setLowerBound:(numValue + 1) upperBound:-1];
	else if(0 == strcmp(ineqValue, ">="))
		[self setLowerBound:numValue upperBound:-1];
	else if(0 == strcmp(ineqValue, "<"))
		[self setLowerBound:-1 upperBound:(numValue - 1)];
	else if(0 == strcmp(ineqValue, "<="))
		[self setLowerBound:-1 upperBound:numValue];
	else [self setLowerBound:numValue upperBound:numValue];

	// Do standard finalization.
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];

	// Strip down to first parser arg.
	[self removeNonParsing];
	for(i=(int)[self count]-1; i>0; i--) [self removeObjectAtIndex:i];
	if([self count] < 1) { [self release]; return nil; }
	else return [self autorelease];
}
- (elt)evalWith:state
{
	// Build a generic list containing the repetitions.
	// Check bounds.
	int n;
	id ctor = [self defaultConstructorWith:state];
	id value = [ctor new];
	id input = [state input];
	elt mark = [input nextKey];
	for(n=0; upperBound<0 || n<upperBound; n++)
	{
		id arg = [[self objectAtIndex:0] evalWith:state].id_u;
		if(arg) [value appendObject:arg];
		else break;
	}
	if(!value || (lowerBound>0 && n<lowerBound))
		{ [value release]; [input setNextKey:mark]; return ELT(id, nil); }
	else return ELT(id, [[self traceObject:value with:state] autorelease]);
}
- initWithLowerBound:(int)lb upperBound:(int)ub
{
	if(![self initWithType:@encode(id) capacity:1]) return nil;
	[self setLowerBound:lb upperBound:ub];
	return self;
}
- init
{
	return [self initWithLowerBound:(-1) upperBound:(-1)];
}
- (void)setLowerBound:(int)lb upperBound:(int)ub
{
	lowerBound = lb; upperBound = ub;
}
- (void)getLowerBound:(int *)lbp upperBound:(int *)ubp
{
	if(lbp) *lbp = lowerBound;
	if(ubp) *ubp = upperBound;
}
- printIvarsTo:stream
{
	[super printIvarsTo:stream];
	if(lowerBound >= 0) [stream writeFormat:" %d...", lowerBound];
	else [stream writeFormat:" ..."];
	if(upperBound >= 0) [stream writeFormat:"%d", upperBound];
	return self;
}
@end

@implementation ParserComposition
- (BOOL)isFinalized
{
	if(!function) return NO;
	return [super isFinalized];
}
- finalizeWith:state
{
	// Extract first function syntax arg as head.  Sole remaining
	// function syntax arg is generic list of real args; merge
	// real arg list into ourselves.
	id replacement, h;
	if([self isFinalized]) return self;
	if(!(h = [[self removeObjectAtIndex:0] finalizeWith:state])) return nil;
	[self setFunction:h];
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	[self removeNonParsing];
	if([self count]==1 && [[self objectAtIndex:0] conformsToProtocol:@protocol(Parsing)])
	{
		[self appendContentsOf:[self removeObjectAtIndex:0]];
		[self removeNonParsing];
	}
	if(![self function]) { [self release]; return nil; }
	return [self autorelease];
}
- (elt)evalWith:state
{
	id value;
	
	[[state push] putObject:self atName:"$"];
	value = [[function evalWith:state].id_u retain];
	[state pop];
	return ELT(id, [[self traceObject:value with:state] autorelease]);
}
- printIvarsTo:stream
{
	[super printIvarsTo:stream];
	[stream writeFormat:" ->"];
	if(function) [function printLispTo:stream];
	else [stream writeFormat:"()"];
	return self;
}
- function
{
	return function;
}
- setFunction:newFunction
{
	[function release];
	function = [newFunction retain];
	return self;
}
- (void)dealloc
{
	[function release];
	[super dealloc];
}
@end

@implementation ParserAlternation
- (BOOL)isFinalized
{
	int i, n;
	if([self count]<=1 || ![super isFinalized]) return NO;
	n = [self count];
	for(i=0; i<n; i++) if(![[self objectAtIndex:i] conformsToProtocol:@protocol(Parsing)]) return NO;
	return YES;
}
- finalizeWith:state
{
	id replacement;
	
	// Sole func syntax arg is generic list of "|"-sep alternatives;
	// merge list of alternatives into ourselves.  (If more than one
	// function syntax arg, do not merge.)  If only one alternative
	// left, replace ourselves with it.
	if([self isFinalized]) return self;
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	[self removeNonParsing];
	if([self count] == 1)
	{
		[self appendContentsOf:[self removeObjectAtIndex:0]];
		[self removeNonParsing];
	}
	if([self count] == 0) { [self release]; return nil; }
	else if([self count] == 1) return [[self autorelease] removeObjectAtIndex:0];
	else return [self autorelease];
}
- (elt)evalWith:state
{
	id value = nil;
	int i, n = [self count];
	id input = [state input];
	elt mark = [input nextKey];
	for(i=0; i<n; i++)
		if( (value = [[self objectAtIndex:i] evalWith:state].id_u) )
			return ELT(id, [self traceObject:value with:state]);
	[input setNextKey:mark];
	return ELT(id, nil);
}
@end

@implementation ParserDefinition
- (BOOL)isFinalized
{
	return NO;
}
- finalizeWith:state
{
	// Register a parser name binding and die.
	// First syntax arg is symbol being defined, next
	// function is parser which forms its definition.
	// All other syntactical arguments are discarded.
	const char *name = 0;
	id replacement, symbol, func = nil;
	symbol = [[[self removeObjectAtIndex:0] finalizeWith:state] retain];
	if(!(replacement = [super finalizeWith:state])) { [symbol release]; return nil; }
	else if(replacement != self) { [symbol release]; return replacement; }
	else [self retain];
	[self removeNonParsing];
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	//!!! leaking name string since libobjects does not distinguish between strings and atoms
	//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	if(getStringValue(symbol, &name) && name && (name = strdup(name))
		&& (func = [[self removeObjectAtIndex:0] retain]))
			[state putObject:func atName:name];
	[symbol release];
	[self release];
	return [func autorelease];
}
- (elt)evalWith:state
{
	// Exists only to register a parser definition and die.
	return ELT(id, nil);
}
@end

@implementation ParserList
- (BOOL)isFinalized
{
	return NO;
}
- finalizeWith:state
{
	// Flatten syntax of Y-separated list of X, from
	// (X ((Y X) (Y X) ...)) to a generic list (X Y X Y X ...).
	int i, n;
	id ctor, reps, replacement;
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	ctor = [self defaultConstructorWith:state];
	replacement = [ctor new];
	[replacement appendObject:[self objectAtIndex:0]];
	reps = [self objectAtIndex:1];
	n = [reps count];
	for(i=0; i<n; i++) [replacement appendContentsOf:[reps objectAtIndex:i]];
	[self release];
	return [replacement autorelease];
}
- (elt)evalWith:state
{
	// Exists only to flatten list and die.
	return ELT(id, nil);
}
@end

@implementation ParserOption
- (BOOL)isFinalized
{
	if([self count] != 1) return NO;
	return [super isFinalized];
}
- finalizeWith:state
{
	// Remove extra parsers to get exactly one.
	int i;
	id replacement;
	if([self isFinalized]) return self;
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	[self removeNonParsing];
	for(i=(int)[self count]-1; i>0; i--) [self removeObjectAtIndex:i];
	if([self count] < 1) { [self release]; return nil; }
	else return [self autorelease];
}
- (elt)evalWith:state
{
	id arg, ctor = [self defaultConstructorWith:state];
	id value = [ctor new];
	if(!value) return ELT(id, nil);
	if( (arg = [[self objectAtIndex:0] evalWith:state].id_u) )
		{ [value appendObject:arg]; }
	return ELT(id, [[self traceObject:value with:state] autorelease]);
}
@end

@implementation ParserSymbol
- (BOOL)isFinalized
{
	if(!symbolName || [self count]!=0) return NO;
	return YES;
}
- finalizeWith:state
{
	// Discard all arguments, after extracting symbol name from first.
	const char *newName = 0;
	id replacement;
	if([self isFinalized]) return self;
	if(!symbolName && getStringValue([[self removeObjectAtIndex:0] finalizeWith:state], &newName) && newName)
		[self setCStringValue:newName];
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	[self empty];
	if(![self cStringValue]) { [self release]; return nil; }
	else return [self autorelease];
}
- (elt)evalWith:state
{
	if(!symbolName) return nil;
	return ELT(id, [self traceObject:[[state objectAtName:symbolName] evalWith:state].id_u with:state]);
}
- initWithCString:(const char *)symbName
{
	if(![self initWithType:@encode(id) capacity:0]) return nil;
	[self setCStringValue:symbName];
	return self;
}
- (void)dealloc
{
	FREE(symbolName);
	[super dealloc];
}
- deepenCopy
{
	if(![super deepenCopy]) return nil;
	if(symbolName && !(symbolName = strdup(symbolName))) { [self dealloc]; return nil; }
	return self;	
}
- (void)setCStringValue:(const char *)aString
{
	FREE(symbolName);
	if(aString) symbolName = strdup(aString);
}
- (const char *)cStringValue { return symbolName; }
- printIvarsTo:stream
{
	[super printIvarsTo:stream];
	if(symbolName) [stream writeFormat:" ->%s", symbolName];
	else [stream writeFormat:" ->()"];
	return self;
}
@end

@implementation ParserLiteral
- (BOOL)isFinalized
{
	if(!literal || [self count]!=0) return NO;
	return YES;
}
- finalizeWith:state
{
	// Discard all arguments, after extracting literal string from first.
	id replacement;
	const char *litValue = 0;
	if([self isFinalized]) return self;
	if(!literal && getStringValue([[self removeObjectAtIndex:0] finalizeWith:state], &litValue) && litValue)
		[self setCStringValue:litValue];
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	[self empty];
	if(![self cStringValue]) { [self release]; return nil; }
	else return [self autorelease];
}
- (elt)evalWith:state
{
	// Discard any leading ignorables.
	int i, c;
	id value, input = [state input];
	elt mark = [input nextKey];
	[[state objectAtName:"ignorable"] evalWith:state];
	for(i=0, c=[input elementAtNextKey].int_u;
		literal[i] && c!=EOF && c==literal[i];
		i++, c=[input elementAtNextKey].int_u);
	if(literal[i]) { [input setNextKey:mark]; return nil; }
	else if(c != EOF) [input setNextKey:[input currentKey]];
	//!!! Would be nice to atomize this string rather than copying.
	if(!(value = [[PValue alloc] initElement:ELT_STR(literal) type:ENCODE_MUTABLE_STRING]))
		{ [input setNextKey:mark]; return nil; }
	return ELT(id, [[self traceObject:value with:state] autorelease]);
}
- initWithCString:(const char *)literalString
{
	if(![self initWithType:@encode(id) capacity:0]) return nil;
	[self setCStringValue:literalString];
	return self;
}
- (void)dealloc
{
	FREE(literal);
	[super dealloc];
}
- deepenCopy
{
	if(![super deepenCopy]) return nil;
	if(literal && !(literal = strdup(literal))) { [self dealloc]; return nil; }
	return self;	
}
- (void)setCStringValue:(const char *)literalString
{
	FREE(literal);
	if(literalString) literal = strdup(literalString);
}
- (const char *)cStringValue { return literal; }
- printIvarsTo:stream
{
	[super printIvarsTo:stream];
	if(literal) [stream writeFormat:" \"%s\"", literal];
	else [stream writeFormat:" \"\""];
	return self;
}
@end

@implementation ParserArgument
- (BOOL)isFinalized
{
	if(argIndex<0 || [self count]!=0) return NO;
	return YES;
}
- finalizeWith:state
{
	// Discard all arguments, after extracting first numerical arg.
	int idx = (-1), i, n;
	id replacement;
	if([self isFinalized]) return self;
	if(!(replacement = [super finalizeWith:state])) return nil;
	else if(replacement != self) return replacement;
	else [self retain];
	n = [self count];
	for(i=0; i<n; i++) if(getIntValue([self objectAtIndex:i], &idx)) { idx--; break; }
	if(idx >= 0) [self setIntValue:idx];
	[self empty];
	if([self intValue] < 0) { [self release]; return nil; }
	else return [self autorelease];
}
- (elt)evalWith:state
{
	return ELT(id, [self traceObject:[[[state objectAtName:"$"]
		objectAtIndex:argIndex] evalWith:state].id_u with:state]);
}
- init
{
	return [self initWithInt:(-1)];
}
- initWithInt:(int)argIdx
{
	if(![self initWithType:@encode(id) capacity:0]) return nil;
	[self setIntValue:argIdx];
	return self;
}
- initWithType:(const char *)enc capacity:(unsigned)cap
{
	if(![super initWithType:@encode(id) capacity:0]) return nil;
	[self setIntValue:(-1)];
	return self;
}
- (void)setIntValue:(int)argIdx { argIndex = argIdx; }
- (int)intValue { return argIndex; }
- printIvarsTo:stream
{
	[super printIvarsTo:stream];
	if(argIndex >= 0) [stream writeFormat:" $%d", argIndex+1];
	else [stream writeFormat:" $$"];
	return self;
}
@end



