// ParserValuing.h
//
// The <ParserValuing> protocol must be implemented by all
// objects produced as values of functions in the Parser
// package's OOF language.  (Functions can also return
// non-object values.)
//
// Some of the most interesting object types returned from
// functions are functions themselves, as happens for example
// when a parser is being built from a BNF specification by
// the parser parser.  This is reflected by the inclusion of
// -evalWith: in the <ParserValuing> protocol (all object
// values should be able to serve as functions, even if they
// only return themselves!), and also in the incorporation of
// the <ParserValuing> protocol into the <Parsing> protocol for
// honest functions (all functions should be returnable as values).
//
// As part of the BNF parsing process (i.e. the evaluation of
// the parser parser), each object value is given a chance to
// "finalize" itself, in order to transform the pure syntax tree
// into a semantic tree of (object-oriented) function calls.
// This is of course most important for the object values
// which happen to be functions.
//
// Parser values are reference counted.  The -finalizeWith: and
// -evalWith: methods should produce autoreleased values; the
// -finalizeWith: method should assume an autoreleased receiver.
// This refcounting means you do not have to worry about freeing
// the temporaries produced by these operations.  Since
// autoreleased objects are not freed right away (in this case only
// when the current scope is closed -- see <ParserStateCollecting>),
// you still have the chance to -retain them if you really want them.
// Because of this reference counting, you must generally subclass
// -finalizeWith: like this:
//
//   -finalizeWith:state
//   {
//     // Check if we should be doing this.
//     if([self isFinalized]) return self;
//
//     // Grab extra syntactical arguments that cannot be handled by
//     // the superclass (finalize if you consume them).  Enclose code
//     // in -retain/-autorelease pair if autoreleases might be flushed.
//     .... int x = [[[self removeObjectAtIndex:7]
//                          finalizeWith:state] intValue] ....
//
//     // Let the superclass do its thing; new object may result.
//     if(!(replacement = [super finalizeWith:state])) return nil;
//     else if(replacement != self) return replacement;
//
//     // Perform finer semantic interpretation of arguments
//     // already processed by superclass.  Enclose code in
//     // -retain/-autorelease pair if autoreleases might be flushed.
//     .... [[self objectAtIndex:3] setSquoggleFactor:x] ....
//
//     return self;
//   }
//     
//
//----------------------------------------------------------------
// OOF Parser Package
// Copyright 1995 by Paul Burchard
// Distributed according to the GNU Library General Public License
//----------------------------------------------------------------

#import <objects/Retaining.h>
#import <objects/elt.h>

// Element conversion macros guaranteed not have any side effects.
// This implementation uses GNU CC "constructor expressions".
#define ELT(typ, val) ({ elt _elt_v; _elt_v.typ ## _u = (val); _elt_v; })
#define ELT_STR(val) ({ elt _elt_v; _elt_v.char_ptr_u = (val); _elt_v; })

@protocol ParserValuing <Retaining>
- finalizeWith:state;     // value need not be self; return value and incoming self are autoreleased
- (BOOL)isFinalized;      // whether -finalizeWith: is still needed
- (elt)evalWith:state;    // value is autoreleased (if object)
- (const char *)evalType; // elt type of result of -evalWith:
@end


