/* alloc.h
 *
 * Convenient macros for memory allocation in C.  As long as all pointers
 * are initialized to NULL, only the macros ALLOC and FREE are ever needed.
 *
 * The following macros all "return" the new address of the allocated array
 * (or NULL).  They also make sure the requested array length is positive.
 *
 *	ALLOC(x, int, 100)
 *				allocates or reallocates array pointed to
 *				by x as necessary (depending on whether x is
 *				NULL or not), to make it the requested size
 *
 *	FREE(x)
 *				frees array pointed to by x, and sets x to
 *				NULL (does nothing if x was already NULL)
 *
 *	MALLOC(x, int, 100)
 *				points x to a newly allocated array
 *
 *	REALLOC(x, int, 100)
 *				reallocates the array pointed to by x
 *				(here x is required to be non-NULL)
 *
 * The macro MAKE_POSITIVE() is also defined for integer values.
 */

#ifndef ALLOC

#include <stdlib.h>

#define ALLOC(ptr, type, cnt) ((ptr) ? REALLOC(ptr, type, cnt) : MALLOC(ptr, type, cnt))
#define MALLOC(ptr, type, cnt) ((ptr) = (type *)malloc((unsigned int)MAKE_POSITIVE(cnt)*sizeof(type)))

/*!!! GNU realloc causes memory leak here (failed realloc remains unchanged) */
#define REALLOC(ptr, type, cnt) ((ptr) = (type *)realloc((void *)(ptr), (unsigned int)MAKE_POSITIVE(cnt)*sizeof(type)))
#define FREE(ptr) ((ptr) ? (free(ptr), (ptr)=0) : 0)
#define MAKE_POSITIVE(x) ((x)>0 ? (x) : 1)

#endif
